from ..http import SlashHTTP
from ...enums import CommandType, OptionType
from ...tools import try_get
from ..types import (
    SlashCommand, SlashOption, SlashOptionCollection, 
    SlashPermission, SlashSubcommand, format_name
)

import inspect
from typing import List, Dict

__all__ = (
    'SlashBuilder',
)

class Subcommand(SlashSubcommand):
    __slots__ = SlashSubcommand.__slots__ + ("__group__", "build",)
    def __init__(self, callback, name=None, description=None, options=None) -> None:
        super().__init__(callback, base_names=[], name=name, description=description, options=options)
        self.__group__ = getattr(callback, "__group__", None)
        self.build: SlashBuilder = None
    @property
    def group_name(self) -> str:
        """The name of the parent group"""
        return try_get(self.__group__, 0, None)
    @property
    def group_description(self) -> str:
        """The description of the parent group"""
        return try_get(self.__group__, 1, "\u200b")
    @property
    def has_group(self) -> bool:
        """Whether this command has a parent group"""
        return self.__group__ != None
    def to_super_dict(self):
        base = super().to_dict()
        if self.has_group:
            return SlashOption(OptionType.SUB_COMMAND_GROUP, self.group_name, self.group_description, options=[base]).to_dict()
        return base
    async def invoke(self, ctx, *args, **kwargs):
        if self.build != None:
            return await self.callback(self.build, ctx, *args, **kwargs)
        return await self.callback(ctx, *args, **kwargs)
    async def update_id(self):
        id = await super().update_id()
        if self.build != None:
            self.build._id = id
        return id

class SlashBuilder():
    """A Superclass for building custom Slashcommands
    
    
    Usage
    ------

    .. code-block::

        class MyCommand(SlashBuilder):
            def __init__(self):
                super().__init__()
                self.name = "base_name"
                self.description = "This is the base"

            @SlashBuilder.subcommand("sub")
            async def my_sub(self, ctx):
                ...
    """
    def __init__(self, name=None, description=None, guild_ids=None, guild_permissions=None, default_permission=True) -> None:
        self.__sync__ = True
        self.__aliases__ = []
        self.__auto_defer__ = None
        self.__choice_generators__ = {}
        self.command_type = CommandType.Slash
        
        self._id = None # set later
        self._name = None
        self._http: SlashHTTP = None # set later
        self.name = name

        
        self.description: str = description
        """The description of the base slashcommand"""
        self.guild_ids: List[int] = guild_ids
        """The guild ids in which the command is available. 
        If no guild ids are provided, this command will be a global slashcommand"""
        self.guild_permissions: Dict[int, SlashPermission] = guild_permissions
        """The permisions for different guilds"""
        self.default_permission: bool = default_permission
        """Whether this command can be used by default or not"""
        self.permissions = SlashPermission()
        """The current permissions"""
        self._options = SlashOptionCollection()

    def __init_subclass__(cls) -> None:
        cls.command_type = CommandType.Slash
    
    async def invoke(self, *args, **kwargs):
        await self.callback(*args, **kwargs)

    @property
    def options(self) -> SlashOptionCollection:
        """The options for this slashcommand"""
        return self._options
    @options.setter
    def options(self, value):
        self._options = SlashOptionCollection(value)
    async def _fetch_id(self):
        return await SlashCommand._fetch_id(self)
    async def update_id(self):
        return await SlashCommand.update_id(self)

    @property
    def id(self) -> int:
        """The ID of thte slashcommand"""
        return self._id
    @property
    def name(self):
        """The name of the base slashcommamnd"""
        return self._name
    @name.setter
    def name(self, value):
        self._name = format_name(value)

    def get_subcommands(self) -> List[Subcommand]:
        return [x[1] for x in inspect.getmembers(self, predicate=lambda x: isinstance(x, Subcommand))]
    def has_groups(self):
        return all(x.has_group for x in self.get_subcommands())
    def has_subs(self):
        return len(self.get_subcommands()) > 0
    def _subs_to_dict(self):
        _commands: List[Subcommand] = [
            (x.group_name, x) for x in
                [x[1] for x in inspect.getmembers(self, predicate=lambda x: isinstance(x, Subcommand))]
        ]
        if not all(x[1].has_group for x in _commands):
            return [x[1].to_super_dict() for x in _commands]
        commands = {}
        for x in _commands:
            if x[0] in commands:
                commands[x[0]]["options"] += x[1].to_supper_dict()["options"]
                continue
            commands[x[0]] = x[1].to_super_dict()
        return list(commands.values())

    def to_dict(self):
        return {
            "name": self.name,
            "description": self.description,
            "default_permission": self.default_permission,
            "options": self.options.to_dict() if not self.has_subs() else self._subs_to_dict()
        }

    def register(self, slash: SlashCommand):
        if self.has_subs():
            if self.has_groups():
                for x in self.get_subcommands():
                    setattr(getattr(self, f"{x.callback.__name__}"), "build", self)
                    x.base_names = [self.name] + [x.group_name or ()]
                    if slash.subcommands.get(self.name) == None:
                        slash.subcommands[self.name] = {}
                    if slash.subcommands[self.name].get(x.group_name) == None:
                        slash.subcommands[self.name][x.group_name] = {}
                    slash.subcommands[self.name][x.group_name][x.name] = x
            else:
                for x in self.get_subcommands():
                    setattr(getattr(self, f"{x.callback.__name__}"), "build", self)
                    x.base_names = [self.name]
                    if slash.subcommands.get(self.name) == None:
                        slash.subcommands[self.name] = {}
                    slash.subcommands[self.name][x.name] = x
        else:
            slash.commands[self.name] = self
    @property
    def guild_only(self) -> bool:
        """Whether this command is only available in some guilds or is a global command"""
        return SlashCommand.guild_only.getter(self)
    
    @staticmethod
    def subcommand(name=None, description=None, options=None):
        """Decorator to a callback for a subcommand
        
        Parameters
        ----------
        name: :class:`str`, optional
            The name for the subcommand; default None
        description: :class:`str`, optional
            The description for the subcommand; default None
        options: List[:class:`SlashOption`], optional
            Options for the subcoommand; default None
        
        Usage
        ------

        .. code-block::

            @SlashBuilder.subcommand("subcommand", "This is a subcoommand", [SlashOption(...)])
            async def sub_callback(self, ctx, ...):
                ...
        """
        def wrapper(callback) -> Subcommand:
            return Subcommand(callback, name, description, options)
        return wrapper
    @staticmethod
    def group(name, description=None):
        """Decorator for a subcommand group
        
        Parameters
        ----------
        name: :class:`str`
            The name of the group
        description: :class:`str`, optional
            The description of the group; default None


        Usage
        -----

        .. code-block::

            @SlashBuilder.group("group_name")
            @SlashBuilder.subcommand(...)
            async def sub_callback(self, ctx, ...):
                ...
        
        """
        def wrapper(callback):
            callback.__group__ = (format_name(name), description,)
            return callback
        return wrapper


from __future__ import annotations

import functools
import logging
import warnings
from typing import TYPE_CHECKING, Any, Callable, List, TypeVar

__all__ = (
    'components_to_dict',
)


class _All():
    def __init__(self) -> None:
        pass
    def __contains__(self, _):
        return True
    def __iter__(self):
        return iter([True])

All = _All()


class _MISSING:
    def __repr__(self) -> str:
        return "..."
    def __eq__(self, o: object) -> bool:
        return isinstance(o, _MISSING)
    def __ne__(self, o: object) -> bool:
        return not self.__eq__(o)
    def __str__(self) -> str:
        return self.__repr__()
    def __sizeof__(self) -> int:
        return 0
    def __len__(self) -> int:
        return 0
    def __contains__(self, value):
        return False
    def get(self, *args):
        return self
class _EMPTY_CHECK():
    """An empty check that will always return True"""
    def __call__(self, *args: Any, **kwds: Any) -> Any:
        return True
    def __repr__(self) -> str:
        return "empty_check"

MISSING = _MISSING()
EMPTY_CHECK = _EMPTY_CHECK()

R = TypeVar("R")

if TYPE_CHECKING:
    from typing_extensions import ParamSpec
    P = ParamSpec('P')


def deprecated(instead=None) -> Callable[[Callable[P, R]], Callable[P, R]]:
    def wrapper(callback: Callable[P, R]) -> Callable[P, R]:
        @functools.wraps(callback)
        def wrapped(*args: P.args, **kwargs: P.kwargs):
            if instead is not None:
                msg = f"{callback.__name__} is deprecated, use {instead} instead!"
            else:
                msg = f"{callback.__name__} is deprecated!"
            # region warning
            # turn filter off
            warnings.simplefilter('always', DeprecationWarning)
            # warn the user, use stacklevel 2
            warnings.warn(msg, stacklevel=2, category=DeprecationWarning)
            # turn filter on again
            warnings.simplefilter('default', DeprecationWarning) 
            # endregion

            return callback(*args, **kwargs)
        return wrapped
    return wrapper


def _raise(ex):
    """Method to raise an exception in a context where the normal `raise` context cant be used"""
    raise ex
def _none(*args, empty_array=False):
    return all(x in [None, MISSING] + [[], [[]]][empty_array is True] for x in args)
def _or(*args, default=None):
    for i, _ in enumerate(args):
        if not _none(args[i]):
            return args[i]
    return default
def _default(default, *args, empty_array=True):
    if _none(*args, empty_array=empty_array):
        return default
    if len(args) == 1:
        return args[0]
    return args
def try_get(value, index, default):
    try:
        return value[index]
    except (IndexError, TypeError):
        return default

def setattribute(object, attribute, value):
    setattr(object, attribute, value)
    return object

def get_index(l: list, elem: Any, mapping = lambda x: x, default: int = -1) -> int:
    """Returns the index of an element in the list
    
    Parameters
    ----------
    l: :class:`List`
        The list to get from
    elem: :class:`Any`
        The element that has to be found
    mapping: :class:`function`
        A function that will be applied to the current element that is checked, before comparing it to the target
    default: :class:`Any`
        The element that will be returned if nothing is found; default -1

    Returns
    -------
    :class:`Any`
        The found element
    

    Example:
    ```py
    any_list = [(1, 2), (2, 3), (3, 4)]
    get(any_list, 4, lambda x: x[1])
    ```
    """
    i = 0
    for x in l: 
        if mapping(x) == elem: 
            return i
        i += 1
    return default

def get(l: list, elem: Any = True, mapping = lambda x: True, default: Any = None, check=lambda x: True):
    """Gets a element from a list
    
    Parameters
    ----------
    l: :class:`List`
        The list to get from
    elem: :class:`Any`
        The element that has to be found
    mapping: :class:`function`
        A function that will be applied to the current element that is checked, before comparing it to the target
    default: :class:`Any`
        The element that will be returned if nothing is found; default None

    Returns
    -------
    :class:`Any`
        The found element


    Example:
    ```py
    any_list = [(1, 2), (2, 3), (3, 4)]
    get(any_list, 4, lambda x: x[1])
    ```
    """
    for x in l:
        if mapping(x) == elem and check(x) is True:
            return x
    return default

def components_to_dict(components) -> List[dict]:
    """Converts a list of components to a dict that can be used for other extensions
    
    Parameters
    ----------
    components: :class:`list`
        A list of components that should be converted.
        
    Example
    
    .. code-block::

        components_to_dict([Button(...), [LinkButton(...), Button(...)]])

    Raises
    ------
    :class:`Exception`
        Invalid data was passed
    
    Returns
    -------
    List[:class:`dict`]
        The converted data


    """
    wrappers: List[List[Any]] = []
    component_list = []

    if len(components) > 1:
        curWrapper = []
        i = 0   # 
        for component in components:
            # if its a subarray
            if hasattr(component, "items") or isinstance(component, (list, tuple)):
                # if this isnt the first line
                if i > 0 and len(curWrapper) > 0:
                    wrappers.append(curWrapper)
                curWrapper = []
                
                # ActionRow was used
                if hasattr(component, "items"):
                    wrappers.append(component.items)
                # ComponentStore was used
                elif hasattr(component, "_components"):
                    wrappers.append(component._components) 
                else:
                    # just comepletely append the components to all rappers
                    wrappers.append(component)
                continue
                
            # i > 0 => Preventing empty component field when first button wants to newLine 
            if component.component_type == 3:
                if i > 0:
                    wrappers.append(curWrapper)
                curWrapper = []
                wrappers.append(component)
            elif component.component_type == 2:
                if component.new_line and i > 0:
                    wrappers.append(curWrapper)
                    curWrapper = [component]
                else: 
                    curWrapper.append(component)
                    i += 1
        if len(curWrapper) > 0:
            wrappers.append(curWrapper)
    else:
        wrappers = [components]
    
    for wrap in wrappers:
        if isinstance(wrap, list) and not all(hasattr(x, "to_dict") for x in wrap):
            raise Exception("Components with types [" + ', '.join([str(type(x)) for x in wrap]) + "] are missing to_dict() method")
        component_list.append({"type": 1, "components": [x.to_dict() for x in wrap] if type(wrap) is list else [wrap.to_dict()]})
    return component_list

def setup_logger(name):
    """
    Thx redstone ;)
    https://github.com/RedstoneZockt/rotstein-dc-py/blob/main/rotstein_py/logging.py
    """
    level = logging.ERROR

    logger = logging.getLogger(name)
    logger.setLevel(level)
    
    stream_handler = logging.StreamHandler()
    formatter = logging.Formatter(f"%(levelname)s: %(message)s")
    stream_handler.setFormatter(formatter)
    stream_handler.setLevel(level)
    
    logger.addHandler(stream_handler)
    return logger


from __future__ import annotations

import discord

import inspect
from enum import IntEnum
from typing import Union

Channel = Union[
    discord.TextChannel, 
    discord.VoiceChannel, 
    discord.StageChannel,
    discord.StoreChannel,
    discord.CategoryChannel, 
    discord.StageChannel,
]
"""Typing object for all possible channel types, only for type hinting"""

_Mentionable = Union[
    Channel,
    discord.Member,
    discord.Role
]
"""Typing object for possible returned classes in :class:`~OptionType.Mentionable`, only for type hinting"""
class __Mentionable:
    """Empty class for comparing a class to `Mentionable`"""
    pass

Mentionable: _Mentionable = __Mentionable
"""Type for a SlashOption with type of :class:`Mentionable`

Usage:
~~~~~~

.. code-block::

    @ui.slash.command(options=[SlashOption(Mentionable, "an_option")])
    async def test(ctx, an_option):
        ...

or

.. code-block::

    @ui.slash.command()
    async def test(ctx, an_option: Mentionable): # OptionType of 'an_option' is now Mentionable
        ... 
"""

class BaseIntEnum(IntEnum):
    def __str__(self) -> str:
        return self.name


class ButtonStyle(BaseIntEnum):
    Blurple     =     Primary           = 1
    Grey        =     Secondary         = 2
    Green       =     Succes            = 3
    Red         =     Destructive       = 4
    URL         =     Link              = 5

    @classmethod
    def getColor(cls, s):
        if isinstance(s, int):
            return cls(s)
        if isinstance(s, cls):
            return s
        s = s.lower()
        if s in ("blurple", "primary"):
            return cls.Blurple
        if s in ("grey", "gray", "secondary"):
            return cls.Grey
        if s in ("green", "succes"):
            return cls.Green
        if s in ("red", "danger"):
            return cls.Red

class CommandType(BaseIntEnum):
    Slash       =              1
    """Chat-input command"""
    User        =              2
    """User-context command"""
    Message     =              3
    """Message-context command"""

    @staticmethod
    def from_string(typ):
        if isinstance(typ, str):
            if typ.lower() == "slash":
                return CommandType.Slash
            elif typ.lower() == "user":
                return CommandType.User
            elif typ.lower() == "message":
                return CommandType.Message
        elif isinstance(typ, CommandType):
            return typ
        else:
            return CommandType(typ)
    def __str__(self):
        return self.name
class ComponentType(BaseIntEnum):
    Action_row      =           1
    Button          =           2
    Select          =           3
class OptionType(BaseIntEnum):
    Subcommand              =          SUB_COMMAND          =           1
    Subcommand_group        =          SUB_COMMAND_GROUP    =           2
    String                  =          STRING               =           3
    Integer                 =          INTEGER              =           4
    """Any integer between -2^53 and 2^53"""
    Boolean    =   bool     =          BOOLEAN              =  BOOL  =  5
    Member     =   user     =          MEMBER               =  USER  =  6
    Channel                 =          CHANNEL              =           7
    """Includes all channel types + categories"""
    Role                    =          ROLE                 =           8
    Mentionable             =          MENTIONABLE          =           9
    """Includes users and roles"""
    Float      =   Number   =          FLOAT                =  NUMBER = 10
    """Any double between -2^53 and 2^53"""

    @classmethod
    def any_to_type(cls, whatever) -> OptionType:
        """Converts something to a option type if possible"""
        if isinstance(whatever, int) and whatever in range(1, 11):
            return whatever
        if inspect.isclass(whatever):
            if whatever is str:
                return cls.String
            if whatever is int:
                return cls.Integer
            if whatever is bool:
                return cls.Boolean
            if whatever in [discord.User, discord.Member]:
                return cls.Member
            if whatever in [discord.TextChannel, discord.VoiceChannel, discord.StageChannel, discord.CategoryChannel]:
                return cls.Channel
            if whatever is discord.Role:
                return cls.Role
            if whatever is Mentionable:
                return cls.Mentionable
            if whatever is float:
                return cls.Float
        if isinstance(whatever, str):
            whatever = whatever.lower()
            if whatever in ["str", "string", "text", "char[]"]:
                return cls.String
            if whatever in ["int", "integer", "number"]:
                return cls.Integer
            if whatever in ["bool", "boolean"]:
                return cls.Boolean
            if whatever in ["user", "discord.user", "member", "discord.member", "usr", "mbr"]:
                return cls.Member
            if whatever in ["channel"]:
                return cls.Channel
            if whatever in ["role", "discord.role"]:
                return cls.Role
            if whatever in ["mentionable", "mention"]:
                return cls.Mentionable
            if whatever in ["float", "floating", "floating number", "f"]:
                return cls.Float
        if isinstance(whatever, list):
            ret = cls.Channel
            ret.__channel_types__ = whatever
            return ret
        if isinstance(whatever, range):
            if float in [cls.any_to_type(whatever.start), cls.any_to_type(whatever.stop)]:
                _type = cls.Float
            else:
                _type = cls.Integer
            _type.__min__, _type.__max__ = whatever.start, whatever.stop
            return _type
class Limits:
    """Limits for OptionType Parameters"""
    class Numeric:
        min, max = -2**53, 2**53

class InteractionResponseType(BaseIntEnum):
    Pong                        =       1
    """respond to ping"""
    # Ack                         =       2  # deprecated
    # """``deprecated`` acknowledge that message was received"""
    # Channel_message             =       3  # deprecated
    # """``deprecated`` respond with message"""
    Channel_message             =       4
    """
    respond with message
        `command` | `component`"""
    Deferred_channel_message    =       5
    """
    defer interaction
        `command | component`"""
    Deferred_message_update     =       6
    """
    update message later
        `component`"""
    Message_update              =       7
    """
    update message for component
        `component`"""
    Autocomplete_result         =       8
    """
    respond with auto-complete choices
        `command`"""



from .errors import NoCommandFound
from ..tools import get, setup_logger
from ..http import BetterRoute, handle_rate_limit, send_files

from discord.http import HTTPClient
from discord.state import ConnectionState
from discord import NotFound, HTTPException, Forbidden


import aiohttp

logging = setup_logger(__name__)

class SlashHTTP():
    def __init__(self, client) -> None:
        self._http: HTTPClient = client.http
        self.token: str = client._connection.http.token
        self.application_id: int = client.user.id
    async def respond_to(self, interaction_id, interaction_token, response_type, data=None, files=None):
        route = BetterRoute("POST", f'/interactions/{interaction_id}/{interaction_token}/callback')
        payload = {
            "type": getattr(response_type, "value", response_type)
        }
        if data:
            payload["data"] = data
        if files is not None:
            return await send_files(route, files, payload, self._http)
        return await self._http.request(route, json=payload)
    async def fetch_command(self, id, guild_id=None):
        try:
            if guild_id:
                return await self._http.request(BetterRoute("GET",f"/applications/{self.application_id}/guilds/{guild_id}/commands/{id}"))
            return await self._http.request(BetterRoute("GET", f"/applications/{self.application_id}/commands/{id}"))
        except HTTPException as ex:
            if ex.status == 429:
                await handle_rate_limit(await ex.response.json())
                return await self.fetch_command(id, guild_id)
            raise ex
    async def get_command(self, command_name, guild_id=None, type=None):
        return get(
            (
                await self.get_global_commands()
            ) if guild_id is None else (
                await self.get_guild_commands(guild_id)
            ), check=lambda x: x.get("name") == command_name and (type or x.get("type")) == x.get("type")
        )
    async def get_id(self, command_name, guild_id=None, type=None):
        found = await self.get_command(command_name, guild_id, getattr(type, "value", type))
        if found is None:
            raise NoCommandFound("No command found with name '" + command_name + "'")
        return found.get('id')

    async def delete_global_commands(self):
        commands = await self._http.request(BetterRoute("GET", f"/applications/{self.application_id}/commands"))
        for x in commands:
            await self.delete_global_command(x["id"])
    async def delete_guild_commands(self, guild_id):
        try:
            commands = await self._http.request(BetterRoute("GET", f"/applications/{self.application_id}/guilds/{guild_id}/commands"))
            for x in commands:
                await self.delete_guild_command(x["id"], guild_id)
        except Forbidden:
            logging.warn("got forbidden in " + str(guild_id))

    async def delete_global_command(self, command_id):
        try:
            return await self._http.request(BetterRoute("DELETE", f"/applications/{self.application_id}/commands/{command_id}"))
        except HTTPException as ex:
            if ex.status == 429:
                await handle_rate_limit(await ex.response.json())
                return await self.delete_global_command(command_id)
            raise ex
    async def delete_guild_command(self, command_id, guild_id):
        try:
            return await self._http.request(BetterRoute("DELETE", f"/applications/{self.application_id}/guilds/{guild_id}/commands/{command_id}"))
        except HTTPException as ex:
            if ex.status == 429:
                await handle_rate_limit(await ex.response.json())
                return await self.delete_guild_command(command_id, guild_id)
            raise ex

    async def get_command_permissions(self, command_id, guild_id):
        try:
            return await self._http.request(BetterRoute("GET", f"/applications/{self.application_id}/guilds/{guild_id}/commands/{command_id}/permissions"))
        except NotFound:
            return {"id": command_id, "application_id": self.application_id, "permissions": []}
        except HTTPException as ex:
            if ex.status == 429:
                await handle_rate_limit(await ex.response.json())
                return await self.get_command_permissions(command_id, guild_id)
            else:
                raise ex
    async def update_command_permissions(self, guild_id, command_id, permissions):
        async with aiohttp.ClientSession() as client:
            async with client.put(f"https://discord.com/api/v9/applications/{self.application_id}/guilds/{guild_id}/commands/{command_id}/permissions",
                headers={"Authorization": "Bot " + self.token}, json={"permissions": permissions}) as response:
                if response.status == 200:
                    return await response.json()
                elif response.status == 429:
                    data = await handle_rate_limit(await response.json())
                    await self.update_command_permissions(guild_id, command_id, permissions)
                    return data
                raise HTTPException(response, response.content)

    async def create_global_command(self, command: dict):
        try:
            return await self._http.request(BetterRoute("POST", f"/applications/{self.application_id}/commands"), json=command)
        except HTTPException as ex:
            if ex.status == 429:
                await handle_rate_limit(await ex.response.json())
                return await self.create_global_command(command)
            raise ex
    async def create_guild_command(self, command, guild_id, permissions = []):
        try:
            data = await self._http.request(BetterRoute("POST", f"/applications/{self.application_id}/guilds/{guild_id}/commands"), json=command)
            await self.update_command_permissions(guild_id, data["id"], permissions)
            return data
        except HTTPException as ex:
            if ex.status == 429:
                await handle_rate_limit(await ex.response.json())
                return await self.create_guild_command(command, guild_id, permissions)
            raise ex


    async def edit_global_command(self, command_id: str, new_command: dict):
        try:
            return await self._http.request(BetterRoute("PATCH", f"/applications/{self.application_id}/commands/{command_id}"), json=new_command)
        except HTTPException as ex:
            if ex.status == 429:
                await handle_rate_limit(await ex.response.json())
                return await self.edit_global_command(command_id, new_command)
            raise ex
    async def edit_guild_command(self, command_id, guild_id: str, new_command: dict, permissions: dict=None):
        try:
            data = await self._http.request(BetterRoute("PATCH", f"/applications/{self.application_id}/guilds/{guild_id}/commands/{command_id}"), json=new_command)
            if permissions is not None:
                return await self.update_command_permissions(guild_id, data["id"], permissions)
        except HTTPException as ex:
            if ex.status == 429:
                await handle_rate_limit(await ex.response.json())
                return await self.edit_guild_command(command_id, guild_id, new_command, permissions)
            raise ex

    async def get_global_commands(self):
        try:
            return await self._http.request(BetterRoute("GET", f"/applications/{self.application_id}/commands"))
        except HTTPException as ex:
            if ex.status == 429:
                await handle_rate_limit(await ex.response.json())
                return await self.get_global_commands()
            raise ex
    async def get_guild_commands(self, guild_id):
        try:
            return await self._http.request(BetterRoute("GET", f"/applications/{self.application_id}/guilds/{guild_id}/commands"))
        except HTTPException as ex:
            if ex.status == 429:
                await handle_rate_limit(await ex.response.json())
                return await self.get_guild_commands(guild_id)
            if ex.status == 403:
                logging.warning("got forbidden in " + str(guild_id))
                return []
            raise ex

# just for typing
class ModifiedSlashState(ConnectionState):
    slash_http: SlashHTTP = None
    http: HTTPClient
